'''

Wraps and extends the
"Microsoft Active Accessibility
API":http://msdn.microsoft.com/library/default.asp?url=/library/en-us/msaa/msaaccrf_87ja.asp
to allow monitoring and control of Windows programs.

This is a layered implementation.

0 The extension module 'pyAAc' uses "SWIG":http://www.swig.org/ to
provide a low-level interface to the MSAA API through the
"IAccessible":http://msdn.microsoft.com/library/default.asp?url=/library/en-us/msaa/msaaccrf_5q05.asp
object. 'pyAAc' also provides access to the 'SetWinEventHook' mechanism.

1 This module first wraps 'IAccessible' with the 'AccessibleObject'
class which provides a more Pythonic interface but basically the same
functionality. We also wrap the 'SetWinEventHook' system with the
'SetWinEventHook' class that provides for automatic cleanup.

2 On top of this is built 'AAbase' which hides the horrors of
**reentry** in WinEventHooks and provides a mechanism for cleanup.

3 'WindowWatcher' extends 'AAbase' to automate monitoring the coming
and going of windows.

4 'AAapp' is derived from 'AAbase' and specialize it for applications
with *top-level* windows.

5 'AAIE' further specializes 'AAapp' to add facilities for monitoring
and controlling Microsoft Internet Explorer.

'''

from Queue import Queue, Empty
from threading import Lock
from traceback import print_exc

import win32gui
import win32con
import win32com.client
from win32com.client import DispatchWithEvents
WrapDispatch = win32com.client.__WrapDispatch

from pyAAc import *

#
# Provide a mapping from idEvent to the name of the constant.
#
WinEventNames = {
  0x0001: "EVENT_SYSTEM_SOUND",
  0x0002: "EVENT_SYSTEM_ALERT",
  0x0003: "EVENT_SYSTEM_FOREGROUND",
  0x0004: "EVENT_SYSTEM_MENUSTART",
  0x0005: "EVENT_SYSTEM_MENUEND",
  0x0006: "EVENT_SYSTEM_MENUPOPUPSTART",
  0x0007: "EVENT_SYSTEM_MENUPOPUPEND",
  0x0008: "EVENT_SYSTEM_CAPTURESTART",
  0x0009: "EVENT_SYSTEM_CAPTUREEND",
  0x000A: "EVENT_SYSTEM_MOVESIZESTART",
  0x000B: "EVENT_SYSTEM_MOVESIZEEND",
  0x000C: "EVENT_SYSTEM_CONTEXTHELPSTART",
  0x000D: "EVENT_SYSTEM_CONTEXTHELPEND",
  0x000E: "EVENT_SYSTEM_DRAGDROPSTART",
  0x000F: "EVENT_SYSTEM_DRAGDROPEND",
  0x0010: "EVENT_SYSTEM_DIALOGSTART",
  0x0011: "EVENT_SYSTEM_DIALOGEND",
  0x0012: "EVENT_SYSTEM_SCROLLINGSTART",
  0x0013: "EVENT_SYSTEM_SCROLLINGEND",
  0x0014: "EVENT_SYSTEM_SWITCHSTART",
  0x0015: "EVENT_SYSTEM_SWITCHEND",
  0x0016: "EVENT_SYSTEM_MINIMIZESTART",
  0x0017: "EVENT_SYSTEM_MINIMIZEEND",
  0x8000: "EVENT_OBJECT_CREATE",
  0x8001: "EVENT_OBJECT_DESTROY",
  0x8002: "EVENT_OBJECT_SHOW",
  0x8003: "EVENT_OBJECT_HIDE",
  0x8004: "EVENT_OBJECT_REORDER",
  0x8005: "EVENT_OBJECT_FOCUS",
  0x8006: "EVENT_OBJECT_SELECTION",
  0x8007: "EVENT_OBJECT_SELECTIONADD",
  0x8008: "EVENT_OBJECT_SELECTIONREMOVE",
  0x8009: "EVENT_OBJECT_SELECTIONWITHIN",
  0x800A: "EVENT_OBJECT_STATECHANGE",
  0x800B: "EVENT_OBJECT_LOCATIONCHANGE",
  0x800C: "EVENT_OBJECT_NAMECHANGE",
  0x800D: "EVENT_OBJECT_DESCRIPTIONCHANGE",
  0x800E: "EVENT_OBJECT_VALUECHANGE",
  0x800F: "EVENT_OBJECT_PARENTCHANGE",
  0x8010: "EVENT_OBJECT_HELPCHANGE",
  0x8011: "EVENT_OBJECT_DEFACTIONCHANGE",
  0x8012: "EVENT_OBJECT_ACCELERATORCHANGE"
}

#
# Provide a mapping from the ObjId to the type of the object.
#
ObjIdNames = {
  0x00000000: "OBJID_WINDOW",
  0xFFFFFFFF: "OBJID_SYSMENU",
  0xFFFFFFFE: "OBJID_TITLEBAR",
  0xFFFFFFFD: "OBJID_MENU",
  0xFFFFFFFC: "OBJID_CLIENT",
  0xFFFFFFFB: "OBJID_VSCROLL",
  0xFFFFFFFA: "OBJID_HSCROLL",
  0xFFFFFFF9: "OBJID_SIZEGRIP",
  0xFFFFFFF8: "OBJID_CARET",
  0xFFFFFFF7: "OBJID_CURSOR",
  0xFFFFFFF6: "OBJID_ALERT",
  0xFFFFFFF5: "OBJID_SOUND"
}

class AccessibleObject:
  '''Encapsulates IAccessible and the child index into a single object.

  Provides methods comparable to those of IAccessible.

  '''
  def __init__(self, ia=None, child=None):
    '''initialize from the raw materials, only used internally'''
    self.ia = ia
    self.child = child

  def Name(self):
    '''Return the string name of the object.'''
    return Cget_accName(self.ia, self.child)

  def ClassName(self):
    '''Return the window class name of the object.'''
    return Cget_accClass(self.ia)

  def Role(self):
    '''return the int 'role' of the object.'''
    return Cget_accRole(self.ia, self.child)

  def RoleText(self):
    '''Return the text description of the role of the object.'''
    return CGetRoleText(self.Role())

  def State(self):
    '''Return the state as an int bit mask.'''
    return Cget_accState(self.ia, self.child)

  def StateText(self):
    '''Return the text describing the full state.

    Handles all the bits, not just one like the low-level API.'''
    result = ''
    state = self.State()
    for i in range(16): # for each bit in the mask
      mask = 1<<i
      if (state & mask) != 0:
        if result:
          result = result + '+' + CGetStateText(mask)
        else:
          result = CGetStateText(mask)
    return result

  def ChildCount(self):
    '''Return the number of children.'''
    if self.child == CHILDID_SELF:
      return Cget_accChildCount(self.ia)
    else:
      return 0

  def Value(self):
    '''Get the 'Value' of the object as a string.'''
    return Cget_accValue(self.ia, self.child)

  def Parent(self):
    '''Return the 'AccessibleObject' that is the parent.'''
    if self.child == CHILDID_SELF:
      return AccessibleObject(Cget_accParent(self.ia), CHILDID_SELF)
    else:
      return AccessibleObject(self.ia, CHILDID_SELF)

  def DoDefaultAction(self):
    '''Make the AccessibleObject do whatever it does.'''
    return CaccDoDefaultAction(self.ia, self.child)

  def SetFocus(self):
    '''Give this object the focus if possible.'''
    Cset_accFocus(self.ia)

  def GetFocus(self):
    '''Get the object with the focus.'''
    tmp = Cget_accFocus(self.ia)
    if tmp is None:
      return None
    if isinstance(tmp, int):
      return AccessibleObject(self.ia, tmp)
    return AccessibleObject(tmp, CHILDID_SELF)

  def Selection(self):
    '''Return a list of 'AccessibleObject' that are the selected children.'''
    if self.child != CHILDID_SELF:
      return []
    tmp = Cget_accSelection(self.ia)
    if tmp == None:
      return []
    if isinstance(tmp, int):
      return [AccessibleObject(self.ia, tmp)]
    elif isinstance(tmp, list):
      result = []
      for t in tmp:
        if isinstance(t, int):
          result.append(AccessibleObject(self.ia, t))
        else:
          result.append(AccessibleObject(t, CHILDID_SELF))
      return result
    else:
      return AccessibleObject(tmp, CHILDID_SELF)

  def Children(self):
    '''Return the children as a list of 'AccessibleObject'.'''
    if self.child != CHILDID_SELF:
      return []
    children = CAccessibleChildren(self.ia)
    res = []
    for c in children:
      if type(c) == type(1):
        res.append(AccessibleObject(self.ia, c))
      else:
        res.append(AccessibleObject(c, CHILDID_SELF))
    return res

  def Location(self):
    '''Return the screen location of the object as a 4-tuple of int.'''
    return CaccLocation(self.ia, self.child)

  def Select(self, flagsSelect):
    '''Set the selection.'''
    CaccSelect(self.ia, flagsSelect, self.child)

  def HitTest(self, point):
    '''Return the AccessibleObject contained in this one that contains the point.'''
    if ai.child:
      return None
    res = CaccHitTest(self.ia, point)
    if res == None:
      return None
    elif type(res) == type(1):
      return AccessibleObject(self.ia, res)
    else:
      return AccessibleObject(res, CHILDID_SELF)

  def Description(self):
    '''Return the string object description.'''
    return Cget_accDescription(self.ia, self.child)

  def Help(self):
    '''Return the help string for the object.'''
    return Cget_accHelp(self.ia, self.child)

  def KeyboardShortcut(self):
    '''Return the shortcut or mnemonic associated with an object.'''
    return Cget_accKeyboardShortcut(self.ia, self.child)

  def Window(self):
    '''Return the window handle ('hwnd') for the object.'''
    return CWindowFromAccessibleObject(self.ia)

  def FindOneChild(self, predicate):
    '''
    Return an AccessibleObject that satisfies the predicate.

    Parameters:

      predicate -- A function of one argument that returns True for the
      desired object.

    '''
    try:
      if predicate(self): # try the predicate of the argument
        return self
    except Error:
      return None

    try:
      for ac in self.Children(): # otherwise consider the children
        try:
          res = ac.FindOneChild(predicate)
        except:
          continue
        if res is not None:
          return res
    except Error: # ignore errors generated by pyAA
      return None

  def FindAllChildren(self, predicate, result=[]):
    '''
    Return a list of AccessibleObject that satisfies the predicate.

    Parameters:

      predicate -- A function of one argument that returns True for the
      desired object.

    '''
    try:
      if predicate(self):
        result.append(self)
      for ac in self.Children():
        ac.FindAllChildren(predicate, result)
    except Error:
      pass
    return result

  def ComObject(self, clsid, EventClass=None):
    '''Attempt to get the COM interface to the AccessibleObject.'''
    # make sure the clsid is a string
    clsid = str(clsid)
    # attempt to get the COM interface
    com = CComObject(self.ia, clsid)
    # convert it using Mark Hammdond's code
    com = WrapDispatch(com)
    if EventClass:
      com = DispatchWithEvents(com, EventClass)
    return com

def AccessibleObjectFromPoint(point):
  '''Return the 'AccessibleObject' containing the point (a 2-tuple of int).'''
  tmp = CAccessibleObjectFromPoint(point)
  return AccessibleObject(tmp[0], tmp[1])

def AccessibleObjectFromEvent(hwnd, objid, childid):
  '''Return the 'AccessibleObject' associated with the event.'''
  tmp = CAccessibleObjectFromEvent(hwnd, objid, childid)
  return AccessibleObject(tmp[0], tmp[1])

def AccessibleObjectFromWindow(hwnd, objid):
  tmp = CAccessibleObjectFromWindow(hwnd, objid)
  return AccessibleObject(tmp[0], tmp[1])

class SetWinEventHook:
  '''Encapsulate the WindowsSDK SetWinEventHook mechanism.

  This class automates cleanup with the object is destroyed.
  '''
  def __init__(self,
               callback=None,
               event=(EVENT_MIN, EVENT_MAX),
               idProcess=0,
               idThread=0,
               flags=WINEVENT_SKIPOWNPROCESS):
    '''Set a hook function for the indicated windows events.

    Parameters:

      'callback' -- a Python callable that accepts 7 arguments. It
      will be called when an event that satisfies the criteria below
      occurs.

      'event' -- a single int specifying an event or a 2-tuple of int
      specifying a range of events. The default is to accept all
      events.

      'idProcess' -- an int restricting events to those coming from
      a single process. The default is all processes.

      'idThread' -- an int restricting events to those coming from a
      single thread. The default is all threads.

      'flags' -- an int. See the documentation for
      "SetWinEventHook":http://msdn.microsoft.com/library/default.asp?url=/library/en-us/msaa/msaaccrf_55m3.asp
      for details. We can only do callbacks that are *Out of Context*
      so the only legal values for this flag are 0,
      'WINEVENT_SKIPOWNPROCESS', and 'WINEVENT_SKIPOWNTHREAD'.

    The internal implementation of 'CSetWinEventHook' from 'pyAAc'
    holds a reference to the 'callback' so that it will not accidentally
    go away while the hook is still active. This routinely causes
    circular references that prevent objects from getting destroyed
    automatically.

    '''

    self.handle = None
    if isinstance(event, int):
      event = (event,event)
    if callback:
      self.handle = CSetWinEventHook(event[0], event[1], callback,
                                     idProcess, idThread, flags)

  def __del__(self):
    '''Clean up the hook when the object is destroyed.'''
    if self.handle:
      CUnhookWinEvent(self.handle)

def SearchRecursive(ao, predicate):
  try:
    if predicate(ao): # try the predicate of the argument
      return ao
    for ac in ao.Children(): # otherwise consider the children
      res = SearchRecursive(ac, predicate)
      if res is not None:
        return res
  except Error: # ignore errors generated by pyAA
    return None

class WinEvent(object):
  '''Encapsulate all the info we get in a WinEventHook.'''

  def __init__(self, hEvent, event, hwnd, idObject, idChild, idThread, dwmsEventTime, ao=None):
    '''Initialize from the WinEventHook parameters.

    Parameters:

    'hEvent' -- int handle of the WinEventHook, used for dispatch

    'event' -- int id of the WinEvent that caused the callback
    'pyAA.WinEventNames' is useful to convert this id into a
    meaningful string.

    'hwnd' -- Handle of the window associated with the event, useful
    with win32gui.

    'idObject' -- Type of the object causing the event.
    'pyAA.ObjIdNames' is useful to convert this into a meaniful name.

    'idChild' -- Identifies a subobject of the hwnd if any.

    'idThread' -- The numeric thread id associated with the event.

    'dwmsEventTime' -- The time in milliseconds when the event
    occurred.

    'ao' -- Optional AccessibleObject associated with this event.
    Saves a little work later if it is needed.

    '''
    self.hEvent = hEvent
    self.idEvent = event
    self.hwnd = hwnd
    self.idObject = idObject
    self.idChild = idChild
    self.idThread = idThread
    self.time = dwmsEventTime
    self.AO = ao

  def getAccessibleObject(self):
    '''Return the AccessibleObject from the event.'''
    if self.AO is None:
      try:
        self.AO = AccessibleObjectFromEvent(self.hwnd, self.idObject, self.idChild)
      except Error:
        pass
    return self.AO

  def getEventName(self):
    '''Return the name of the event as a string.'''
    return WinEventNames.get(self.idEvent, hex(self.idEvent))

  def getObjectName(self):
    '''Return the object type as a string.'''
    return ObjIdNames.get(self.idObject, hex(self.idObject))

  ao = property(fget=getAccessibleObject)
  event_name = property(fget=getEventName)
  object_name = property(fget=getObjectName)

  def __str__(self):
    '''Convert to a string.'''
    return ('WinEvent hEvent=%d event=%s hwnd=%x idObject=%s child=%d thread=%d time=%d' %
            (self.hEvent, self.getEventName(), self.hwnd,
             self.getObjectId(), self.idChild, self.idThread,
             self.time))

#
# module variables, kept here in an attempt to avoid circular links
#
_event_queue = Queue() # events to handle later
_reenter_lock = Lock() # a barrier to tell if we're reentered
_all_lock = Lock()     # prevent events from getting left behind in the queue
_event_dispatch_map = {}   # map event handle into Python callback

def event_dispatch(hEvent, event, hwnd, idObject, idChild, idThread, dwmsEventTime):
  '''Central dispatch for WinEvents.

  This is *really* tricky, be careful making changes here. This
  function might be re-entered **MANY** times before it exits. It is
  crtical that this re-entry be properly managed. I handle re-entry
  with a queue of events to handle later and a simple lock. On
  re-entry, the event will simply be queued and the callback will
  immediately return. Only the first call gets to proceed past the
  lock. It will process all the queued entrys before it returns.

  The parameters are defined by the Windows SetWinEventHook
  mechanism. You must accept exactly 7 ints or chaos ensues.

  Parameters:

    'hEvent' -- int handle of the WinEventHook, used for dispatch

    'event' -- int id of the WinEvent that caused the callback
  'pyAA.WinEventNames' is useful to convert this id into a
  meaningful string.

    'hwnd' -- Handle of the window associated with the event, useful
  with win32gui.

    'idObject' -- Type of the object causing the event.
  'pyAA.ObjIdNames' is useful to convert this into a meaniful name.

    'idChild' -- Identifies a subobject of the hwnd if any.

    'idThread' -- The numeric thread id associated with the event.

    'dwmsEventTime' -- The time in milliseconds when the event
  occurred.

  '''

  # construct an event object from the parameters
  event = WinEvent(hEvent, event, hwnd, idObject, idChild, idThread, dwmsEventTime)
  # add it to the queue
  _event_queue.put(event)

  # check to see if this is a reentrant call
  _all_lock.acquire()
  if not _reenter_lock.acquire(0):
    _all_lock.release()
    # we're nested, the event is already queued, just return
    return
  _all_lock.release()

  # we only get this far if we are not nested
  while 1:
    # get an event to process, the queue may have entries added
    # while we work
    try:
      _all_lock.acquire()
      event = _event_queue.get_nowait()
    except Empty:
      # queue is empty, we're done
      _reenter_lock.release()
      _all_lock.release()
      return
    _all_lock.release()

    # process it, printing an exception if any occurs
    try:
      handle = event.hEvent
      disp_func, disp_hwnd, disp_idObject = _event_dispatch_map[handle]
    except KeyError:
      continue # handler is now gone, we don't care anymore

    # filter on the hwnd if requested
    if ((disp_hwnd is None or disp_hwnd == event.hwnd) and
        (disp_idObject is None or disp_idObject == event.idObject)):
      try:
        disp_func(event)
      except:
        print_exc()

def AddWinEventHook(callback=None, event=(EVENT_MIN, EVENT_MAX), idThread=0, hwnd=None,
                    idObject=None):
  '''Add a hook to be invoked when an event matching the criteria occurs.

  Returns an int handle that can be used to delete the hook later.

  Parameters:

    'callback' -- The function to invoke to handle the event.

    'event' -- Restrict the handler to the indicated event (single
    int) or a range of events (2-tuple of ints).

    'idThread' -- Only call the hook for events from this thread.

    'hwnd' -- Only call the hook for events originating from this hwnd.

    'idObject' -- Only call the hook for events on objects with this
    id.

  '''
  # convert single event into a range
  if isinstance(event, int):
    event = (event, event)

  handle = CSetWinEventHook(event[0], event[1], event_dispatch, 0, idThread,
                            WINEVENT_SKIPOWNPROCESS)
  # is it possible for a callback to occur before the next statement? if so we'll need
  # some kind of lock to prevent it
  _event_dispatch_map[handle] = (callback, hwnd, idObject)
  return handle

def DeleteWinEventHook(handle):
  '''Disable a previously established hook.

  Parameters:

  'handle' -- int handle for the callback returned earlier by AddHook.

  '''
  CUnhookWinEvent(handle)
  del _event_dispatch_map[handle]

class AAbase(object):
  '''Base types that provides a mechanism for cleanup when the object is finalized.'''

  def __init__(self):
    '''Initialize the list of hooks to be called on Release.'''
    self.release_hooks = []

  def Release(self):
    '''Run any cleanup hooks that have been requested.'''
    hooks = self.release_hooks
    self.release_hooks = [] # avoid a recursive call
    for hook, hook_args in hooks:
      if hook is not None:
        hook(*hook_args)

  def AddReleaseHook(self, hook, *args):
    '''Add a function to be called on Release.'''
    self.release_hooks.append((hook, args))

  def AddWinEventHook(self, callback=None, event=(EVENT_MIN, EVENT_MAX), idThread=0,
                      hwnd=None, idObject=None):
    '''Add a WinEventHook and a ReleaseHook to clean it up.'''
    handle = AddWinEventHook(callback=callback, event=event, idThread=idThread,
                             hwnd=hwnd, idObject=idObject)
    self.AddReleaseHook(DeleteWinEventHook, handle)

class AAproxy(object):
  '''Avoid circular references by creating a proxy object.

  The proxy acts in every way like the original but it is a separate object. When you
  drop the last reference to the proxy, the original will be cleaned up.

  '''

  def __init__(self, obj):
    '''Add the object to this instances dictionary.'''
    self.__dict__['_obj_'] = obj

  def __del__(self):
    '''When the proxy is deleted inform the original.'''
    self._obj_.Release()

  def __getattr__(self,name):
    '''Map attributes from the original.'''
    return getattr(self._obj_, name)

  def __setattr__(self,name,value):
    '''Map attributes from the original.'''
    return setattr(self._obj_, name, value)

class WindowWatcher(AAbase):
  '''Watch for windows being created.

  It is expected that the user will set the instance variable 'client_classes'
  to a list of classes to try on every new window. We will attempt to construct
  an instance of each class in 'client_classes' until one succeeds. The class
  constructor should raise 'TypeError' if they fail.

  Public methods:

    'CloseAllClients' -- Tell each of the client windows to close.

  Instance variables (all private):

    'clients' -- map from threadIds to an object derived from AAbase.

    'client_classes' -- list of classes that are tried sequentially
    until one successfully complete construction. On failure the class
    should raise 'TypeError'.

  '''

  def __init__(self, client_classes = []):
    '''Intialize the instance variables.'''

    AAbase.__init__(self)

    self.clients = {} # client windows we're observing
    self.client_classes = client_classes # a list of classes to test new windows

    self.AddWinEventHook(callback=self.handleShowEvent, event=EVENT_OBJECT_SHOW,
                         idObject=OBJID_WINDOW)

  def handleShowEvent(self, event):
    '''Watch for window creates.

    Parameters:

    'event' -- 'WinEvent' object with the info on what happened,
    where, and when.

    '''

    try:
      ao = event.getAccessibleObject();
      if ao is None:
        return

      if event.idThread in self.clients:
        # we've already got this one
        return

      # invoke each of the class filters to see if any like this object
      ao = event.getAccessibleObject()
      name = ao.Name()
      for client_class in self.client_classes:
        try:
          obj = client_class(event)
        except TypeError:
          continue

        obj = AAproxy(obj) # convert to a proxy to avoid circular problems
        self.clients[event.idThread] = obj
        obj.AddReleaseHook(self.releaseClient, event.idThread)
        obj.OnStart(event)
        break
    except Error:
      return

  def releaseClient(self, idThread):
    '''Drop our reference to the client.

    This is the callback given to clients we are observing so they can
    tell us when they quit on their own.

    '''
    del self.clients[idThread]

  def CloseAllClients(self):
    '''Send the close message to all our clients. For testing. '''
    for client in self.clients.values():
      client.Close()

class AAapp(AAbase):
  '''ActiveAccessibility interface to applications.

  Public methods:

  Close -- send a close message to the window.

  OnStart -- called by the WindowWatcher when it starts managing a window.

  OnDestroy -- called when the top-level window is destroyed.

  '''

  def __init__(self, event):
    '''Attempt to initialize the interface to the window.

    Raise TypeError if construction fails.

    Parameters:

      'event' -- the 'WinEvent' object that brought us here.

    '''

    AAbase.__init__(self)

    # climb to the toplevel window
    top = event.getAccessibleObject()
    parent = top.Parent()
    while parent.Name() != 'Desktop':
      top = parent
      parent = top.Parent()

    self.top = top

    # watch for a destory event on the top window
    self.AddWinEventHook(callback = self.handleDestroyEvent,
                         event = EVENT_OBJECT_DESTROY,
                         idThread = event.idThread,
                         hwnd = self.top.Window())

  def handleDestroyEvent(self, event):
    '''Callback to cleanup on destruction of the window.

    Parameters:

      'event' -- 'WinEvent' object.

    '''
    self.OnDestroy(event) # let a derived class know if it cares
    self.Release()

  def Close(self):
    '''Tell the window to close.'''
    hwnd = self.top.Window()
    win32gui.PostMessage(hwnd, win32con.WM_CANCELMODE, 0, 0)
    win32gui.PostMessage(hwnd, win32con.WM_CLOSE, 0, 0)

  def OnStart(self, event):
    '''Notify me when the application starts.'''
    pass

  def OnDestroy(self, event):
    '''Notify me when the application has closed.'''
    pass

class AAIE(AAapp):
  '''ActiveAccessibility interface to Internet Explorer.

  Provides methods to monitor page loading, browser creation, and browser destruction events.

  Properties:

  'address' -- the URL as a string. When you set it the browser will
  navigate to the new address.
  'images' -- a list of all the accessibility objects for images on the page. Read only.

  Public methods:

  'OnAddressChange' -- called when the text in the address box changes.
  'OnPageDone' -- called when a page is finished loading.
  '''

  def __init__(self, event):
    '''Initialize the instance variables for IE.'''

    # filter on the name before trying to construct the base class
    ao = event.getAccessibleObject()
    name = ao.Name()
    if not name or name.find('Microsoft Internet') < 0:
      raise TypeError

    AAapp.__init__(self, event)
    #get the address object
    self.address_object = self.top.FindOneChild(lambda ao: (ao.Role() == ROLE_SYSTEM_TEXT and
                                                            ao.Name() == 'Address'))
    if self.address_object is None:
      raise TypeError

    #get the statusbar object
    self.status_object = self.top.FindOneChild(lambda ao: (ao.Parent().Role()
                                                           == ROLE_SYSTEM_STATUSBAR))
    if self.status_object is None:
      raise TypeError

    #get the pane parent
    pane = self.top.FindOneChild(lambda ao: ao.Role() == ROLE_SYSTEM_PANE)
    self.pane_parent = pane.Parent()

    #watch for changes in the status bar
    self.AddWinEventHook(callback = self.StatusbarChange, event = EVENT_OBJECT_NAMECHANGE,
                            hwnd = self.status_object.Window(), idThread = event.idThread)

    #watch for changes in the address
    self.AddWinEventHook(callback = self.AddressChange, event = EVENT_OBJECT_VALUECHANGE,
                         hwnd = self.address_object.Window(), idThread = event.idThread)

    self.ready = False

  def getImages(self):
    '''Return all of the images on the page.'''
    return self.pane_parent.FindAllChildren(lambda ao: (ao.Role() == ROLE_SYSTEM_GRAPHIC), [])

  def getAddress(self):
    '''Return the current browser URL as a string.'''
    return self.address_object.Value()

  def setAddress(self, addr):
    '''Tell the browser to navigate to the URL.'''
    hwnd = self.AddressObject.Window()
    win32gui.SendMessage(hwnd, win32con.WM_SETTEXT, 0, addr)
    win32gui.SendMessage(hwnd, win32con.WM_KEYDOWN, win32con.VK_RETURN, 1)

  #
  # Properties
  #
  address = property(fget=getAddress, fset=setAddress)
  images = property(fget=getImages)

  def AddressChange(self, event):
    '''Notify internally when the address changes.

    Parameters:

    'event' -- the 'WinEvent' object for the event.
    '''
    self.ready = True
    self.OnAddressChange(event)

  def StatusbarChange(self, event):
    '''Notify internally when the first text box in the status bar changes.

    Parameters:

    'event' -- the 'WinEvent' object for the event.
    '''
    try:
      n = self.status_object.Name()
    except Error:
      return

    #figure out if the page is done loading
    if n == 'Done' and self.ready == True:
      try:
        self.pane_object.Name()
      except:
        self.ready = False
        self.OnPageDone(event)

  def OnPageDone(self, event):
    '''A place holder method. Override this to receive page done events.'''
    pass

  def OnAddressChange(self, event):
    '''A place holder method. Override this to receive address change events.'''
    pass

if __name__ == '__main__':
  # code for testing
  from wxPython.wx import *
  import webbrowser, sys

  class IEEvents:
    '''You can define events here but do not have to.'''
    def Ononblur(self):
      print 'blur'

  class myAAIE(AAIE):
    '''Specialize the IE interface to display navigation.'''

    def __init__(self, event):
      '''Tack on some COM stuff to demonstrate what we can do.'''
      AAIE.__init__(self, event)

      self.com = self.pane_parent.ComObject('{332C4427-26CB-11D0-B483-00C04FD90119}', IEEvents)

      # I can add events here and give them names I like!
      self.com.Ononfocus = self.OnFocus
      self.com.Ononload = self.OnLoad

      self.AddReleaseHook(self.Cleanup)

    def Cleanup(self):
      print 'cleanup'
      del self.com

    def OnFocus(self):
      print 'focus'

    def OnLoad(self):
      print 'onload', self.address

    def OnPageDone(self, event):
      print 'Page finished loading:', self.address

    def OnStart(self, event):
      print 'Browser started', self.address

    def OnDestroy(self, event):
      print 'Browser destroyed'

  # sequence through these URLs for testing
  URLs = [ 'http://www.cs.unc.edu',
           'http://www.cs.unc.edu/~gb',
           'http://www.cs.unc.edu/~parente',
           'http://wwwx.cs.unc.edu/Research/assist/et/',
           'http://www.cs.unc.edu/Research/assist/',
           'http://www.cs.unc.edu/Research/assist/people.shtml',
           'http://www.cs.unc.edu/~gb/office.htm' ]

  class myTextCtrl(wxTextCtrl):
    '''Specialize wxTextCtrl to enable redirecting stdout.'''
    def write(self, str):
      self.AppendText(str)

  class myFrame(wxFrame):
    '''A simple frame to display our text.'''

    def __init__(self):
      '''Do the layout.'''
      id = wxNewId()
      wxFrame.__init__(self, None, id, "Try WinEvents", size=wxSize(900, 600))
      self.text = myTextCtrl(self, -1, style = wxTE_READONLY | wxTE_MULTILINE | wxTE_RICH)
      self.text.SetFont(wxFont(10, wxMODERN, wxNORMAL, wxNORMAL))
      #sys.stdout = self.text

      self.chook = AAproxy(WindowWatcher([myAAIE]))
      EVT_CLOSE(self, self.OnClose)

      if len(sys.argv) > 1:
        # if an argument is provided that create browsers automatically
        duration = int(sys.argv[1])

        id = wxNewId()
        self.timer = wxTimer(self, id)
        EVT_TIMER(self, id, self.OnTimer)
        self.step = 0 # just an index to modulate the activity

        print 'testing for %d steps' % duration
        self.max_step = duration

        self.timer.Start(1000)

    def OnTimer(self, event):
      '''make things happen'''
      self.step += 1

      if self.step == self.max_step:
        self.chook.CloseAllClients()

      elif self.step > self.max_step:
        self.timer.Stop()
        self.Close()

      elif self.step % 4 == 0:
        self.chook.CloseAllClients()

      else:
        url = URLs[self.step % len(URLs)]
        print 'start', url
        webbrowser.open_new(url)

    def OnClose(self, evt):
      '''clean up when the window is closed'''
#      self.chook.CloseAllClients()
      evt.Skip()

  class myApp(wxApp):
    def OnInit(self):
      self.frame = myFrame()
      self.frame.Show(True)
      self.SetTopWindow(self.frame)
      return true

  # finally run the demo
  app = myApp(0)
  app.MainLoop()
